OAREPO_VALIDATE_MERGER = 'oarepo_validate.utils.merge'
"""
Merger function for merging partially validated data into stored record.
Must have a signature ``merger(record: Record, data: dict): None`` and
merge the data into the record.

Properties in the record not present in data should be left intact.
"""



#
# Copyright (c) 2019 UCT Prague.
#
# record.py is part of Invenio Explicit ACLs
# (see https://github.com/oarepo/invenio-explicit-acls).
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""ACL related mixins for Record class."""
import json
from functools import wraps

from invenio_jsonschemas import current_jsonschemas
from invenio_records import Record
from jsonpatch import apply_patch


def convert_relative_schema_to_absolute(x):
    """Convert relative record schema to absolute if needed."""
    if x.startswith('http://') or x.startswith('https://'):
        return x
    return current_jsonschemas.path_to_url(x)


class AllowedSchemaMixin(object):
    """A mixin that keeps allowed and preferred schema. Not to be used directly."""

    # DO NOT forget to set these up in subclasses
    ALLOWED_SCHEMAS = ()
    PREFERRED_SCHEMA = None
    _RESOLVED = False

    @classmethod
    def _prepare_schemas(cls):
        """Converts ALLOWED_SCHEMAS and PREFERRED_SCHEMA to absolute urls."""
        if not cls._RESOLVED:
            cls.ALLOWED_SCHEMAS = tuple(convert_relative_schema_to_absolute(x) for x in cls.ALLOWED_SCHEMAS)
            cls.PREFERRED_SCHEMA = convert_relative_schema_to_absolute(cls.PREFERRED_SCHEMA)
            cls._RESOLVED = True

    @classmethod
    def _convert_and_get_schema(cls, data):
        """Locate $schema in data and if needed convert it to absolute. Returns the converted schema."""
        cls._prepare_schemas()
        schema = data.get('$schema')
        if schema:
            absolute_schema = convert_relative_schema_to_absolute(schema)
            if absolute_schema is not None and schema != absolute_schema:
                schema = absolute_schema
                data['$schema'] = absolute_schema
        return schema


class SchemaKeepingRecordMixin(AllowedSchemaMixin):
    """
    A mixin for Record class that makes sure $schema is always in allowed schemas.

    Note that this mixin is not enough, always use invenio_explicit_acls.marshmallow.SchemaEnforcingMixin
    as well. The reason is that Invenio does not inject custom Record implementation for PUT, PATCH and DELETE
    operations.
    """

    # DO NOT forget to set these up in subclasses
    ALLOWED_SCHEMAS = ()
    PREFERRED_SCHEMA = None
    _RESOLVED = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if '$schema' not in self:
            self._prepare_schemas()
            self['$schema'] = self.PREFERRED_SCHEMA
        else:
            self._check_schema(self)

    def clear(self):
        """Preserves the schema even if the record is cleared and all metadata wiped out."""
        schema = self.get('$schema')
        super().clear()
        if schema:
            self['$schema'] = schema

    def update(self, e=None, **f):
        """Dictionary update."""
        self._check_schema(e or f)
        return super().update(e, **f)

    @classmethod
    def _check_schema(cls, data):
        schema = cls._convert_and_get_schema(data)
        if schema:
            if schema not in cls.ALLOWED_SCHEMAS:
                raise AttributeError('Schema %s not in allowed schemas %s' % (data['$schema'], cls.ALLOWED_SCHEMAS))

    def __setitem__(self, key, value):
        """Dict's setitem."""
        if key == '$schema':
            self._prepare_schemas()
            if value not in self.ALLOWED_SCHEMAS:
                value = current_jsonschemas.path_to_url(value)
                if value not in self.ALLOWED_SCHEMAS:
                    raise AttributeError('Schema %s not in allowed schemas %s' % (value, self.ALLOWED_SCHEMAS))
            value = convert_relative_schema_to_absolute(value)
        return super().__setitem__(key, value)

    def __delitem__(self, key):
        """Dict's delitem."""
        if key == '$schema':
            raise AttributeError('Schema can not be deleted')
        return super().__delitem__(key)

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        """
        Creates a new record instance and store it in the database.

        For parameters see :py:class:invenio_records.api.Record
        """
        cls._prepare_schemas()

        if '$schema' not in data:
            data['$schema'] = convert_relative_schema_to_absolute(cls.PREFERRED_SCHEMA)
        else:
            cls._check_schema(data)
        ret = super().create(data, id_, **kwargs)
        return ret

    def patch(self, patch):
        """Patch record metadata. Overrides invenio patch to check if schema has changed

        :params patch: Dictionary of record metadata.
        :returns: A new :class:`Record` instance.
        """
        data = apply_patch(dict(self), patch)
        self._check_schema(data)
        return self.__class__(data, model=self.model)


def files_keeping_wrapper(f):
    @wraps(f)
    def wrap(self, *args, **kwargs):
        bucket_id = self.get('_bucket')
        files = self.get('_files')
        ret = f(self, *args, **kwargs)
        if bucket_id:
            self['_bucket'] = bucket_id
        if files is not None:
            self['_files'] = files
        return ret

    return wrap


def json_equals(a, b):
    a = json.dumps(a, sort_keys=True)
    b = json.dumps(b, sort_keys=True)
    return a == b


class FilesKeepingRecordMixin:

    @files_keeping_wrapper
    def clear(self):
        super().clear()

    @files_keeping_wrapper
    def update(self, *args, **kwargs):
        """Dictionary update."""
        return super().update(*args, **kwargs)

    def patch(self, patch):
        bucket_id = self.get('_bucket')
        files = self.get('_files')

        data = apply_patch(dict(self), patch)
        ret = self.__class__(data, model=self.model)
        if ret.get('_bucket') != bucket_id:
            raise AttributeError('_bucket can not be overwritten')
        if not json_equals(ret.get('_files'), files):
            raise AttributeError('_files can not be overwritten')
        return ret


class SchemaEnforcingRecord(SchemaKeepingRecordMixin, Record):
    ALLOWED_SCHEMAS = ('records/record-v1.0.0.json',)
    PREFERRED_SCHEMA = 'records/record-v1.0.0.json'



import copy
import json
from typing import Iterable

import click
from flask import current_app
from flask.cli import with_appcontext
from invenio_base.utils import obj_or_import_string
from invenio_pidstore.models import PersistentIdentifier, PIDStatus
from invenio_records import Record


@click.group(name='oarepo:validate')
def validate():
    """OARepo record validation commands."""


@validate.command('revalidate')
@click.option('--pid-type', '-t', help='Limit revalidate to a given pid type')
@click.option('--pid', '-p', help='Limit revalidate to a given pid of form pid_type:pid_value')
@click.option('--save/--no-save', '-s', default=False, help='If the validation is successful, commit the record')
@click.option('--verbose/--quiet', '-v', default=False, help='Print details')
@with_appcontext
def import_taxonomy(pid_type=None, pid=None, save=False, verbose=False):
    pids: Iterable[PersistentIdentifier]
    if pid_type:
        pids = PersistentIdentifier.query.filter(
            PersistentIdentifier.pid_type == pid_type,
            PersistentIdentifier.status == PIDStatus.REGISTERED.value)
    elif pid:
        pid_type, pid_value = pid.split(':', maxsplit=1)
        pids = PersistentIdentifier.query.filter(
            PersistentIdentifier.pid_type == pid_type,
            PersistentIdentifier.pid_value == pid_value,
            PersistentIdentifier.status == PIDStatus.REGISTERED.value
        )
    else:
        pids = PersistentIdentifier.query.filter(
            PersistentIdentifier.status == PIDStatus.REGISTERED.value)

    pid_type_to_record_class = {}
    for rec in current_app.config.get('RECORDS_REST_ENDPOINTS', {}).values():
        pid_type_to_record_class[rec['pid_type']] = obj_or_import_string(rec.get('record_class', Record))

    for pid in pids:
        if pid.pid_type not in pid_type_to_record_class:
            if verbose:
                print('Skipping pid', pid)
        if verbose:
            print('Processing pid', pid)
        record = pid_type_to_record_class[pid.pid_type].get_record(pid.object_uuid)
        try:
            data_before = copy.deepcopy(dict(record))
            record.validate()
            data_after = copy.deepcopy(dict(record))
            equals = json.dumps(data_before, sort_keys=True) == json.dumps(data_after, sort_keys=True)
            if verbose:
                print('    valid, content unchanged after validation: %s' % equals)

            if save and not equals:
                if verbose:
                    print('    saving pid', pid)
                record.commit()
        except Exception as e:
            if verbose:
                print('    INVALID, exception', e)



from invenio_base.utils import obj_or_import_string
from werkzeug.utils import cached_property
from . import config


class OARepoValidateState:
    """State for record references."""

    def __init__(self, app):
        """Initialize state."""
        self.app = app

    @cached_property
    def merge_function(self):
        return obj_or_import_string(self.app.config['OAREPO_VALIDATE_MERGER'])


class OARepoValidate:
    def __init__(self, app=None, db=None):
        self.init_app(app, db)

    def init_app(self, app, db):
        self.init_config(app, db)
        state = OARepoValidateState(app)
        app.extensions['oarepo-validate'] = state

    def init_config(self, app, db):
        app.config.setdefault('OAREPO_VALIDATE_MERGER', config.OAREPO_VALIDATE_MERGER)



from invenio_records import Record
from invenio_records_rest.loaders.marshmallow import MarshmallowErrors
from jsonpatch import apply_patch
from marshmallow import ValidationError
from sqlalchemy.util import symbol

from .proxies import current_validate
from .signals import before_marshmallow_validate, after_marshmallow_validate

KNOWN_RECORD_VALIDATION_PROPS = {
    'format_checker',
    'validator'
}


class Keep:
    def __init__(self, value):
        self.value = value


DELETED = symbol('deleted')


class MarshmallowValidatedRecordMixin:
    """
    A mixin that keeps marshmallow schema and PID fetcher. This way later in processing
    (for example, out of invenio REST methods) we can use marshmallow for validation.
    """
    MARSHMALLOW_SCHEMA = None
    """The metadata schema"""

    PID_FETCHER = None
    """
    Optional PID fetcher to set up the persistent identifier in context
    (and have the same context as invenio has)
    """

    VALIDATE_MARSHMALLOW = True
    """
    Setting this variable to True will make the validate() method (called inside commit)
    to trigger marshmallow validation. If the same marshmallow schema is used on rest loaders,
    it will be validated twice.

    To fix this and be safe on REST side, set VALIDATE_MARSHMALLOW to False and VALIDATE_PATCH to True
    """

    MERGE_WITH_VALIDATED = True
    """
    If set to True, if validation is successful the validated data will be merged into the record's
    metadata.
    """

    MERGE_WITH_VALIDATED_ERROR = True
    """
    If set to True, if validation is not successful but partially valid content is reported from
    marshmallow, it will be merged into the record's metadata.
    """

    VALIDATE_PATCH = False
    """
    If VALIDATE_MARSHMALLOW is set to False, setting this variable will cause patch method to perform
    marshmallow validation. See the readme for details
    """

    def validate_marshmallow(self, data=None, validate_kwargs=None):
        """
        Validates marshmallow and returns validated data.
        Does not modify the record nor save it to the database.
        """
        if data is None:
            data = self
        validate_kwargs = validate_kwargs or {}
        context = {**validate_kwargs}
        if self.PID_FETCHER is not None:
            pid = self.__class__.PID_FETCHER(None, data)
            context['pid'] = pid
        context['record'] = self
        before_marshmallow_validate.send(
            self,
            record=self, context=context, **validate_kwargs)
        try:
            result = self.MARSHMALLOW_SCHEMA(context=context).load(data)
            after_marshmallow_validate.send(
                self,
                record=self, context=context, result=result, error=None, **validate_kwargs)
            return result
        except ValidationError as error:
            after_marshmallow_validate.send(
                self,
                record=self, context=context, result=None, error=error, **validate_kwargs)
            err = MarshmallowErrors(error.messages)
            err.valid_data = error.valid_data
            raise err
        except Exception as error:
            after_marshmallow_validate.send(
                self,
                record=self, context=context, result=None, error=error, **validate_kwargs)
            raise

    def patch(self, patch):
        """Patch record metadata. Overrides invenio patch to perform marshmallow validation

        :params patch: Dictionary of record metadata.
        :returns: A new :class:`Record` instance.
        """
        data = apply_patch(dict(self), patch)
        if self.VALIDATE_PATCH:
            data = self.validate_marshmallow(data)
        return self.__class__(data, model=self.model)

    def validate(self, **kwargs):
        """
        Overloaded invenio validate. If VALIDATE_MARSHMALLOW is set on the instance
        or validate_marshmallow parameter set to True, marhsmallow validation will be performed
        as well

        :param validate_marshmallow set to True to perform marshmallow validation. If set to False,
        class-wide VALIDATE_MARSHMALLOW is used.
        """

        if kwargs.pop('validate_marshmallow', self.VALIDATE_MARSHMALLOW):
            try:
                data = self.validate_marshmallow(validate_kwargs=kwargs)
                if self.MERGE_WITH_VALIDATED:
                    self._merge_in(data)
            except MarshmallowErrors as e:
                if self.MERGE_WITH_VALIDATED_ERROR:
                    self._merge_in(e.valid_data)
                raise

        kwargs = {
            k: v for k, v in kwargs.items() if k in KNOWN_RECORD_VALIDATION_PROPS
        }
        return super().validate(**kwargs)

    def _merge_in(self, valid_data):
        current_validate.merge_function(self, valid_data)


class MarshmallowValidatedRecord(MarshmallowValidatedRecordMixin, Record):
    """
    Marshmallow-enabled record
    """
    pass



from blinker import signal

before_marshmallow_validate = signal('oarepo_before_marshmallow_validate')
"""
Signal invoked before record metadata are validated (loaded by marshmallow schema)
inside Record.validate

:param source:  the record being validated
:param record:  the record being validated
:param context: marshmallow context
:param **kwargs: kwargs passed to Record.create or Record.commit (or Record.validate)
"""

after_marshmallow_validate = signal('oarepo_after_marshmallow_validate')
"""
Signal invoked after record metadata are validated (loaded by marshmallow schema)
inside Record.validate

:param source:  the record being validated
:param record:  the record that was successfully validated
:param context: marshmallow context
:param result:  result of load that will be used to update record's metadata.
                Signal handler can modify it.
:param **kwargs: kwargs passed to Record.create or Record.commit (or Record.validate)
"""




from __future__ import absolute_import, print_function

from flask import current_app
from werkzeug.local import LocalProxy

current_validate = LocalProxy(
    lambda: current_app.extensions['oarepo-validate'])
"""Helper proxy to access oarepo validate state object."""



from oarepo_validate.marshmallow import Keep, DELETED


def merge(base, next):
    if next is DELETED:
        return DELETED
    if isinstance(next, Keep):
        return next.value  # perform no merging here
    if isinstance(base, list) or isinstance(next, list):
        return merge_lists(base, next)
    if isinstance(base, dict) or isinstance(next, dict):
        return merge_dicts(base, next)
    return next


def merge_lists(base, next):
    if next is None:
        next = []

    if not isinstance(base, list):
        base = []

    for idx in range(min(len(base), len(next))):
        base[idx] = merge(base[idx], next[idx])
    if len(next) > len(base):
        base.extend(next[len(base):len(next)])
    for k in reversed(range(len(base))):
        if base[k] is DELETED:
            del base[k]
    return base


def merge_dicts(base, next):
    if next is None:
        next = {}
    if not isinstance(base, dict):
        base = {}

    for k, v in list(base.items()):
        if k in next:
            merged = merge(base[k], next[k])
            if merged is DELETED:
                del base[k]
            else:
                base[k] = merged
    for k, v in next.items():
        if k not in base and v is not DELETED:
            base[k] = v
    return base



from flask import request


def json_loader():
    """Do-nothing loader."""
    return request.get_json(force=True)


def json_files_loader():
    """A loader that removes record-files stuff"""
    resp = json_loader()
    if '_bucket' in resp:
        del resp['_bucket']
    if '_files' in resp:
        del resp['_files']
    return resp



from invenio_records_rest.serializers import record_responsify, search_responsify
from invenio_records_rest.serializers.base import TransformerMixinInterface, PreprocessorMixin
from invenio_records_rest.serializers.json import JSONSerializerMixin


class JSONSerializer(JSONSerializerMixin, TransformerMixinInterface, PreprocessorMixin):
    def transform_record(self, pid, record, links_factory=None, **kwargs):
        """Transform record into an intermediate representation."""
        ret = self.preprocess_record(pid, record, links_factory=links_factory, **kwargs)
        if 'pid' in ret:
            ret['id'] = ret.pop('pid').pid_value
        return ret

    def transform_search_hit(self, pid, record_hit, links_factory=None,
                             **kwargs):
        """Transform search result hit into an intermediate representation."""
        ret = self.preprocess_search_hit(pid, record_hit, links_factory=links_factory, **kwargs)
        if 'pid' in ret:
            ret['id'] = ret.pop('pid').pid_value
        if 'highlight' in record_hit:
            ret['highlight'] = record_hit['highlight']
        return ret


json_serializer = JSONSerializer(replace_refs=False)

json_response = record_responsify(json_serializer, 'application/json')
json_search = search_responsify(json_serializer, 'application/json')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 UCT Prague.
#
# oarepo-validate is free software; you can redistribute it and/or modify it under
# the terms of the MIT License; see LICENSE file for more details.

"""OArepo Validate library for record metadata validation"""

import os

from setuptools import find_packages, setup

readme = open('README.md').read()
history = open('CHANGES.md').read()
OAREPO_VERSION = os.environ.get('OAREPO_VERSION', '3.3.0')

install_requires = [
    'wrapt>=1.11.2',
]

tests_require = [
    f'oarepo[tests]~={OAREPO_VERSION}'
]

extras_require = {
    'tests': tests_require,
    'dev': [
        *tests_require,
        'invenio-app',
        'redis',
        'markdown-toc'
    ]
}

setup_requires = [
    'pytest-runner>=2.7',
]

packages = find_packages(exclude=["tests"])

# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('oarepo_validate', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name='oarepo-validate',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    keywords='metadata validation oarepo records',
    license='MIT',
    author='UCT Prague',
    author_email='miroslav.simek@vscht.cz',
    url='https://github.com/oarepo/oarepo-validate',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    entry_points={
        'flask.commands': [
            'oarepo:validate = oarepo_validate.cli:validate',
        ],
        'invenio_base.apps': [
            'oarepo_validate = oarepo_validate.ext:OARepoValidate',
        ],
        'invenio_base.api_apps': [
            'oarepo_validate = oarepo_validate.ext:OARepoValidate',
        ],
    },
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Development Status :: 4 - Beta',
    ],
)



from .record import SchemaKeepingRecordMixin, SchemaEnforcingRecord, FilesKeepingRecordMixin
from .marshmallow import MarshmallowValidatedRecordMixin, MarshmallowValidatedRecord, Keep, DELETED
from .signals import *
from .loaders import json_loader, json_files_loader
from .serializers import JSONSerializer, json_search, json_response
